## Hierarchical clustering using 4N
## version 0.24 / March 2015 for "Towards a hierarchical strategy for finding protein complexes in multi scale IP/MS data"
## (c) 2015/2016, Joachim Kutzera, j.kutzera@uva.nl

fourN = function(IPmat, U=NA, S=NA, C=90, P=0.5, join=T, join_jaccard=F, verbose=F, coOcc = NA){
    ## the whole 4N-procedure in one function
    NNNs = getNNNs(IPmat, U=U, C=C, verbose=verbose, coOcc = coOcc)
    coreComplexes = getCoreComplexes(NNNs$NNN, S=S, verbose=verbose)
    if (join) {
	JC = joinCC(coreComplexes, ovT=P, verbose=verbose, jaccard=join_jaccard)
	return(list(JC=JC, complexes = coreComplexes, U=NNNs$U))
	}else{
	return(list(JC=NA, complexes = coreComplexes, U=NNNs$U))
	}
    }
    
realSimilarity = function(coOcc,U){
    ## for all proteins in the co-occurrence matrix coOcc
    ## at which co-occurrence threshold do all proteins still belong to one connected component in a cluster? this is U
    ## Following from that U: what is the minimum reciprocity level where all components stay together? this is S, not used yet
    if (all(coOcc == 1)) return(list(U=1, S=1))
    coOccs = sort(unique(as.numeric(coOcc)))
    coOccs = coOccs[which(coOccs >= U)]
    a = 0
    oneComponent = T
    coOcc2=NA
    coOcc2_old=NA
    while(oneComponent){
	a = a + 1
	coOcc2_old = coOcc2
	coOcc2 = coOcc
	coOcc2[which(coOcc2 < coOccs[a])] = 0
	coOcc2[which(coOcc2 > 0)] = 1
	G2 = new("graphAM", adjMat = coOcc2)
	components = connectedComp(as(G2, "graphNEL"))
	if (length(components)>1) oneComponent = F
	}
    return(list(U=round(coOccs[a-1],2), S= mean(rowSums(coOcc2_old)-1) / nrow(coOcc)))
    }
    
HC4N = function(IPmat, joined=NA, startU=NA, startS=NA, P=0.5, S = NA, verbose=T, veryverbose=F, C=90){
    ## Every instance of hC4N writes into the same globlal flat list. The hierarchy comes from the "comesFromLevel" vector
    HC4Nenv  <- new.env(parent=.GlobalEnv)
    HC4Nenv$HC4Nres = list()
    
    if (is.na(joined[[1]][1])) {## Level 1 can be done by hand and HC4N works on that. If not done yet, HC4N creates the first level itself
	cat("\nFirst Level:\n")
	fourNres = fourN(IPmat, U=startU, S=startS, P=P, C=C, verbose=T)
	joined = fourNres$JC$joined
	startU = fourNres$U
	}
    cat("\nLevel 2 and above\n")
    recursiveHC4N = function(HC4Nenv, IPmat, joined, usedU, P,comeFrom=NULL){
	rnames = rownames(IPmat)
	for (a in 1:length(joined)){
	    IPsubset = IPmat[joined[[a]],]
	    keep = which(colSums(IPsubset, na.rm=T)>0)
	    if (length(joined[[a]]) > 3){ ## we need more than 3 proteins to continue clustering
		if (length(keep)<=1){ ## the IPsubset only has one column
		    HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = list(joined = joined[[a]], IPsubset=IPsubset, comeFrom = c(comeFrom, a), U = 1, paramU=1, leaf="all prots in one column")
		    }else{ ## more than 3 proteins in more than one column - the normal case
		    IPsubset = IPsubset[,keep]
		    coo = miniCoOcc(IPsubset)
		    NNNs = getNNNs(IPsubset, coOcc = coo, C=C, verbose=veryverbose)
		    CCs = getCoreComplexes(NNNs$NNNs, verbose=veryverbose, S = S, keepAll = T)
		    if (verbose) cat("\ncomeFrom:", c(comeFrom, a), " U ", NNNs$U)
		    nJC = joinCC(CCs, ovT=P, jaccard=T, verbose=veryverbose)
		    realVals = realSimilarity(coo, usedU)
		    realU = realVals$U 
		    ## given the case that one complex contains all proteins in IPmat, we would get into a endless recursion.
		    ## we habe to remove those JCs.
		    newJoined = list()
		    for (b in 1:length(nJC$joined))
			if (length(nJC$joined[[b]]) < nrow(IPsubset))
			    newJoined[[length(newJoined)+1]] = nJC$joined[[b]]
		    ## We need to decide how to process the current set of JCs (nJC): 
		      ## only one complex:
		    if  (length(nJC$joined)==1) {
			HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = 
			list(joined = joined[[a]], IPsubset=IPsubset, comeFrom = c(comeFrom, a), U = realU, paramU = usedU, leaf="all proteins in one cluster")
			}else{
		      ## we can split further:	
			HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = list(joined = joined[[a]], IPsubset=IPsubset, comeFrom = c(comeFrom, a), U = realU, paramU = usedU,leaf="no")
			recursiveHC4N(HC4Nenv, IPsubset, newJoined, usedU = NNNs$U, P = P, comeFrom = c(comeFrom, a))
			}
		    }
		}else{
		## only 2-3 proteins in dataset
		    mc = miniCoOcc(IPsubset)
		    if (nrow(mc) == 2) {
			U = mc[1,2]
			HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = list(joined = joined[[a]], IPsubset=IPsubset, 
			comeFrom = c(comeFrom, a), U = U, paramU = U, leaf="2 proteins left")
			}
		    if (nrow(mc) == 3) {
			U = round(sort(c(mc[1,2],mc[1,3], mc[2,3]))[2],2) ## the min U that keeps the triangle connected
			HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = list(joined = joined[[a]], IPsubset=IPsubset, 
			comeFrom = c(comeFrom, a), U = U, paramU = U, leaf="no")
			## the two most connected proteins get a new own child cluster while the third one is lost
			whichones = list(c(1,2),c(1,3),c(2,3))[[which.max(c(mc[1,2],mc[1,3], mc[2,3]))]]
			HC4Nenv$HC4Nres[[length(HC4Nenv$HC4Nres)+1]] = list(joined = whichones, IPsubset=IPsubset[whichones,], 
			comeFrom = c(comeFrom, a,1), U = round(max(c(mc[1,2],mc[1,3], mc[2,3])),2), paramU = U, leaf="2 proteins of 3 left")
			}
		}	    
	    }
	}
    recursiveHC4N(HC4Nenv, IPmat, joined, usedU=startU, P=P) ## Starting of the recursive 4N application
    ## result processing
    isLeaf = NA
    comesFromRoot = NA
    comesFromLevel = NA
    hasParamU = NA
    hasRealU = NA
    leaf=NA
    for (a in 1:length(HC4Nenv$HC4Nres)){
	if (length(HC4Nenv$HC4Nres[[a]]$comeFrom)==1) comesFromRoot[a] = T else comesFromRoot[a] = FALSE
	comesFromLevel[a] = length(HC4Nenv$HC4Nres[[a]]$comeFrom)
	hasParamU[a] = HC4Nenv$HC4Nres[[a]]$paramU
	hasRealU[a] = HC4Nenv$HC4Nres[[a]]$U
	if (HC4Nenv$HC4Nres[[a]]$leaf=="no") isLeaf[a] = F else isLeaf[a] = T
	leaf[a] = HC4Nenv$HC4Nres[[a]]$leaf
	}
    cat("\n")
    return(list(HC4Nres = HC4Nenv$HC4Nres, isLeaf = isLeaf, comesFromRoot = comesFromRoot, 
    comesFromLevel = comesFromLevel, hasU = hasRealU, paramU = hasParamU, leaf=leaf))
    }
        
getParent = function(HC4Nres, whichOne){ 
    ## gives back the index of the parent of a leaf
    comeFrom = HC4Nres[[whichOne]]$comeFrom
    if (length(comeFrom)==1) return(0)
    b = 1
    notFound=TRUE
    while (notFound){
	if (length(HC4Nres[[b]]$comeFrom) == length(comeFrom)-1)
	    if (all(HC4Nres[[b]]$comeFrom == comeFrom[1:(length(comeFrom)-1)]))
		notFound=F
	b = b + 1
	}
    return (b-1)
    }

getJC = function(rnames, HC4Nres, whichOne=c(1:length(HC4Nres))){
    ## extracts a set of complexes from the HCresult, removing the tree information
    if (class(rnames)=="matrix") stop("Error: rnames must only contain the rownames.")
    JC = list()
    JC$joined = list()
    for (a in 1:length(whichOne)){
	cat(a, " of ", length(whichOne), "\r")
	comeFrom = HC4Nres[[whichOne[a]]]$comeFrom
	if (length(comeFrom)==1) JC$joined[[length(JC$joined)+1]] = HC4Nres[[whichOne[a]]]$joined
	if (length(comeFrom)>1){
	    parent = getParent(HC4Nres, whichOne[a])
	    JC$joined[[length(JC$joined)+1]] = which(rnames %in% rownames(HC4Nres[[parent]]$IPsubset)[HC4Nres[[whichOne[a]]]$joined])
	    }
	}
    JC
    }   

hierarchicalClusterPlot = function(IPmat, HC4Nres, rnames = rownames(IPmat),startU, useHM=F, resort=T, sqrtHM=F){
    ## useHM heatmaps the matrix, sqrtHM heatmaps the squareroot of it.
    prots = NULL
    lv1s = which(HC4Nres$comesFromLevel == 1)
    JC = getJC(rownames(IPmat), HC4Nres$HC4Nres, lv1s)
    for (a in 1:length(JC$joined))
	prots = c(prots, JC$joined[[a]])
    prots = unique(prots)
    HCplot = matrix(nrow=length(prots), ncol=length(prots), data=0)
    rownames(HCplot) = colnames(HCplot) = rnames[prots]
    for (a in 1:length(JC$joined)){
	if (HC4Nres$isLeaf[lv1s[a]])
	    HCplot[rnames[JC$joined[[a]]], rnames[JC$joined[[a]]]] = pmax(HCplot[rnames[JC$joined[[a]]], rnames[JC$joined[[a]]]], HC4Nres$hasU[lv1s[a]])
	    else HCplot[rnames[JC$joined[[a]]], rnames[JC$joined[[a]]]] = pmax(HCplot[rnames[JC$joined[[a]]], rnames[JC$joined[[a]]]], startU)
	}
    ## now: going level by level through the tree
    mlevel = max(HC4Nres$comesFromLevel)
    for (a in 2:mlevel){ 
	whichJCs = which(HC4Nres$comesFromLevel == a)
	whichDads = NA
	for (b in 1:length(whichJCs))
	    whichDads[b] = getParent(HC4Nres$HC4Nres, whichJCs[b])
	uWhichDads = unique(whichDads)
	jc_lastlevel = getJC(rownames(IPmat), HC4Nres$HC4Nres, uWhichDads)
	jc = getJC(rownames(IPmat), HC4Nres$HC4Nres, whichJCs)
	for (b in 1:length(whichJCs))
	    HCplot[rnames[jc$joined[[b]]],rnames[jc$joined[[b]]]][
			      which(HCplot[rnames[jc$joined[[b]]],rnames[jc$joined[[b]]]] < HC4Nres$hasU[whichJCs[b]]) ## leafs get their own U
			      ] = HC4Nres$hasU[whichJCs[b]] ## only old values that are lower get overwritten
	## resorting the parent-blocks, so that the just filled in children blocks look nice
	## makes sense if there's not too much overlap
	if (resort){
	    for (b in 1:length(jc_lastlevel$joined)){
		subSet = rnames[jc_lastlevel$joined[[b]]]
		if (length(subSet)>1){
		    prots_jc = NULL
		    for (c in 1:length(jc$joined))
			if (whichDads[c] == uWhichDads[b])
			    prots_jc = c(prots_jc, rnames[jc$joined[[c]]])
		    prots_jc = unique(prots_jc)
		    prots_jc = prots_jc[which(prots_jc %in% subSet)]
		    notInjc = rnames[jc_lastlevel$joined[[b]]][which(! subSet %in% prots_jc)]
		    if (!is.null(prots_jc)) {## some last level jcs have no children
			HCinds = NA
			for (c in 1:length(subSet))
			    HCinds[c] = which(rownames(HCplot)==subSet[c])
			trnames = rownames(HCplot)
			trnames[HCinds] = c(prots_jc, notInjc)
			HCplot = HCplot[trnames, trnames]
			}
		    }
		}
	    }
	}
    if (useHM) {
	if (!sqrtHM) hm = heatmap_noplot(HCplot)
	else hm = heatmap_noplot(sqrt(HCplot))
	HCplot = HCplot[hm$rowInd, hm$colInd]
	}
    return(HCplot)
    }
      
plotHCplot = function(hcPlot, whichPart=1:nrow(hcPlot), labelType=1, stepDiff=0, xOffset=0, yOffset=0, textCex=0.5, legendCex=0.5, useSqrt = F, maintitle="Hierarchical cluster plot"){
    ## plots the image and labels.
    ## stepDiff: needed sometimes to stretch the labels to fit to the image
    ## yoffset can be set to change the label starting point a bit, so that all labels are readable
    ## labelType: 0 = none, 1 = all labels right next to the image, 2 = labels diagonally across the whole image
    ## whichPart: takes out a small subset from the HCplot and plots it instead
    ## very dark plots are bad for distinguishing low U and no U. sqrt can be set T then.
    hcPlot = hcPlot[whichPart,whichPart]
    srnames = rownames(hcPlot)
    if (useSqrt) hcPlot = sqrt(hcPlot)
    if (!any(c(0,1,2) == labelType)) stop("Please set labelType to 0 for no label, 1 for labels right next to the image and to 2 for a diagonal labeling.")
    fac = 1/nrow(hcPlot)
    l=40
    dif=1.7
    bluegreys2=NA ## for plotting
    bluegreys2[1] = rgb(0,0,0)
    for (a in 2:(l-1)) bluegreys2[a] = rgb(red=((a+1)/(l*dif)),green=((a+1)/(l*dif)),blue=((a+1)/l), max=1)
    bluegreys2[l] = "white"
    l=10
    dif=1.7 ## blueishness
    bluegreys=NA ## for legend (as 40 dots would be too much)
    bluegreys[1] = rgb(0,0,0)
    dscale = NA
    for (a in 2:(l-1)) dscale[a] = ((a+1)/l)
    if (useSqrt)
	dscale = sqrt(dscale)
    for (a in 2:(l-1)) bluegreys[a] = rgb(red=(dscale[a]/dif),green=(dscale[a]/dif),blue=dscale[a], max=1)
    bluegreys[l] = "white"

    fac = 1/nrow(hcPlot) + stepDiff
    if (labelType==0) {
        image(hcPlot, xaxt="n", yaxt="n", col=bluegreys2, cex=1,cex.lab=1.5, xlim=c(0,1), ylim=c(0,1), zlim=c(0,1), main=maintitle)
        }
    if (labelType==1) {
        image(hcPlot, xaxt="n", yaxt="n", col=bluegreys2, cex=1,cex.lab=1.5, xlim=c(0,1.3), ylim=c(0,1), zlim=c(0,1), main=maintitle)
        for (a in 1:length(srnames)) 
            if (a%%2==0) {
                text(1.1,yOffset+(a*fac)-0.5*fac,paste(a,srnames[a]), cex=textCex, col="black") 
                }else{
                text(1.2,yOffset+(a*fac)-0.5*fac,paste(a,srnames[a]), cex=textCex, col="black")
                }
        }
    if (labelType==2) {
        image(hcPlot, xaxt="n", yaxt="n", col=bluegreys2, cex=1,cex.lab=1.5, xlim=c(0,1.05), ylim=c(0,1), zlim=c(0,1), main=maintitle)
        
        for (a in 1:length(srnames)) 
            if (a%%2==0) {
                text(xOffset + (a*fac)-0.5*fac,yOffset+(a*fac)-0.5*fac,paste(a,srnames[a]), cex=textCex, col="red") 
                }else{
                text(xOffset + (a*fac)-0.5*fac,yOffset+(a*fac)-0.5*fac,paste(a,srnames[a]), cex=textCex, col="red")
                }           
        }
    
    legend(0.86, 0.40, round(seq(0,1,1/9),2), fill=bluegreys,bg="white", cex=legendCex)
    }   
   
cosineDist = function(d1, d2){
    div1 = crossprod(d1, d2)
    div2 = sqrt(crossprod(d1) * crossprod(d2))
    if (div1 > div2) div1 = div2
    return(acos( div1 / div2 ) * (180 / pi))
    }
   
getNNNs = function(data, U=NA, C = 90, coOcc=NA, verbose=T){
    ## first step of 4N. if parameter U is NA, it is set automatically.
    if (any(data==0, na.rm=T)) data[which(data==0)] = NA
    NNN = list()
    if (is.na(coOcc[1])) coOcc = miniCoOcc(data, verbose=verbose)
    if (is.na(U)){
	maxCoOcc=NA
	for (a in 1:nrow(coOcc))
	    maxCoOcc[a] = max(coOcc[a,][-a])
	U = min(maxCoOcc[which(maxCoOcc > 0)])
	if (verbose) cat("automatic U: ", U, "\n")
	}
    for (a in 1:nrow(coOcc)){
	NNN[[a]] = list()
	NNN[[a]]$NNNs = which(coOcc[,a]>=U)
	if (length(NNN[[a]]$NNNs)<=1) NNN[[a]]$NNNs = NA
	cat(a, " of ", nrow(coOcc), "\r")
	}
    lost = FALSE
    if (C < 90){
	for (a in 1:nrow(coOcc))
	    if (!is.na(NNN[[a]]$NNNs[1])){
		newNNN = NNN[[a]]$NNNs
		kickOut=rep(FALSE, length(newNNN))
		for (b in 1:length(NNN[[a]]$NNNs)){
		    both = which(!is.na(data[a,]) & !is.na(data[NNN[[a]]$NNNs[b],]))
		    if (length(both) > 1){
			d1 = data[a,both]    
			d2 = data[NNN[[a]]$NNNs[b],both]
			if (cosineDist(d1,d2) > C) kickOut[b] = TRUE
			}
		    }
		    if (length(which(!kickOut))<=1){
			NNN[[a]]$NNNs = NA 
			cat("\nWarning: Proteins got lost due to the cosine restriction. \n")
			}
		    else 
			NNN[[a]]$NNNs = NNN[[a]]$NNNs[! kickOut]
		}
	}
    return(list(NNNs = NNN, U = U))
    }    
    
getCoreComplexes = function(NNNs, S=NA, verbose=T, keepAll = FALSE){
    ## finding the core complex for one single NNN
    ## if S == NA, it gets set autmatically
    ## if S is given and keepAll = true, S is sought first and the given S usd when the auto-S is larger
    if (is.na(S)){#
	autoCCs = findParameterS(NNNs)
	if (verbose) cat("automatic S: ", autoCCs$auto_S, "\n")
	return(autoCCs$CCs)
	}
    if (keepAll) {
	autoCCs = findParameterS(NNNs)
	if (S >= autoCCs$auto_S){
	    if (verbose) cat("automatic S: ", autoCCs$auto_S, "\n")
	    return(autoCCs$CCs)
	    }
	}
    CCs = list()
    for (a in 1:length(NNNs)){
	if (verbose) cat(a, "\r")
	NNN = NNNs[[a]]$NNNs
	if (is.null(NNN) || is.na(NNN)) {
	    CCs[[a]] = NA
	    }else{
	    kickout = TRUE
	    while(any(kickout)){
		kickout=NA
		for (b in 1:length(NNN)){
		    cNNN = NNNs[[NNN[b]]]$NNNs ## reciproge NNN for NNN-prot a of seed
		    coNNN = which(cNNN %in% NNN) ## these reciproge NNN-proteins are also in the NNN of the seed
		    if (length(coNNN) >= (length(NNN)*S)) ## enough NNNs of the seed also occur in the NNNs of this prot
			kickout[b] = FALSE else kickout[b] = TRUE
		    }
		NNN = NNN[!kickout]
		if (all(kickout)) kickout=FALSE ## nothing to kick out anymore.
		}
	    if (length(NNN)>1) CCs[[a]] = NNN else CCs[[a]] = NA
	    }
	}
    return(CCs)
    }
    
miniCoOcc = function(data, tresh=0, verbose=F){
    ## returns the binary jaccard-coefficient for all proteins to all proteins
    ## this is used by getNNNcoOccValue as the coOccs don't need to be calculated over and over again
    coOccs = matrix(nrow=nrow(data),ncol=nrow(data),data=0)
    if (any(is.na(data)))
        data[which(is.na(data))] = 0
    rownames(coOccs)=colnames(coOccs)=rownames(data)
    for (a in 1:(nrow(data)-1)){
        if (verbose) cat(a,"of",nrow(data),"    \r")
        for (b in (a+1):nrow(data)){
            avals = which(data[a,] > tresh)
            bvals = which(data[b,] > tresh)
            occBoth = length(which(avals %in% bvals))
            occA = length(avals) - occBoth
            occB = length(bvals) - occBoth
            if ((occBoth + occA + occB) == 0) stop("you have either all-zero rows or columns!")
            coOccs[a,b] = coOccs[b,a] = occBoth / (occBoth + occA + occB)
            }
        }
    for (a in 1:nrow(data)) coOccs[a,a]=1
    return(coOccs)
    }
    
miniCosineDist = function(data, verbose=F){
    ## returns the cosine distance as angle for all proteins vs each other when they co-occur anywhere
    cosDists = matrix(nrow=nrow(data),ncol=nrow(data),data=90)
    data[which(data==0)] = NA
    rownames(cosDists)=colnames(cosDists)=rownames(data)
    for (a in 1:(nrow(data)-1)){
        if (verbose) cat(a,"of",nrow(data),"    \r")
        for (b in (a+1):nrow(data)){
	    both = which(!is.na(data[a,]) & !is.na(data[b,]))
	    if (length(both) > 0)
		cosDists[a,b] = cosDists[b,a] = cosineDist(data[a,both],data[b,both])
	    }
	for (a in 1:nrow(data)) cosDists[a,a]=1
	}
    return(cosDists)
    }
    
findParameterS = function(NNNs, minS=0.001, steps=40, verbose=F){
     ## finds the highest possible threshold for the CC-jaccard-coefficient that does not cause protein loss
    step = (1-minS) / (steps-1)
    coOccs = seq(minS,1,step)
    if (verbose) cat("\n1\n")
    CCs = getCoreComplexes(NNNs, S=coOccs[1])
    inCC = rep(0, length(NNNs))
        for (b in 1:length(CCs))
            if (!is.null(CCs[[b]][[1]]))
                if (!is.na(CCs[[b]][[1]]))
                    inCC[CCs[[b]]] = 1
    mNumProt = numProt = sum(inCC)
    x=2
    while ((x<=length(coOccs))&&(numProt==mNumProt)){
        if (verbose) cat("\n",x,"\n")
        CCs_old = CCs
        CCs = getCoreComplexes(NNNs, S=coOccs[x])
        inCC = rep(0, length(NNNs))
        for (b in 1:length(CCs))
            if (!is.null(CCs[[b]][[1]]))
                if (!is.na(CCs[[b]][[1]]))
                    inCC[CCs[[b]]] = 1
        numProt = sum(inCC)
        x=x+1
        }
    auto_S = coOccs[x-2]
    return(list(auto_S = auto_S, CCs = CCs_old))
    }

overlap = function(CCs, verbose){
	## calculates the overlap for all complexes vs each other - needed by joinCC
	nnl = length(CCs)
	if (verbose) cat("\nCCs-overlaps: \n")
	overlaps = matrix(nrow=nnl, ncol=nnl, data=0)
		for (a in 1:(nnl-1)) {
		    for (b in (a+1):nnl)
			overlaps[a,b] = length(which(CCs[[a]] %in% CCs[[b]]))
		if (verbose) cat(a, " of ", nnl, "\r")
		}
		overlaps
		}
		
minsAB = function(CCs, verbose){
    ## calculates the length-minimum of all complexes vs each other - needed by joinCC
    minsABs = matrix(nrow=length(CCs), ncol=length(CCs))
    nnl = length(CCs)
    lengths = NA
	for (a in 1:length(CCs))
	    lengths[a] = length(CCs[[a]])
    if (verbose) cat("\nCCs-minimums: \n")
    for (a in 1:(nnl-1)) {
	for (b in (a+1):nnl)
	    minsABs[a,b] = min(lengths[a],lengths[b])
	if (verbose) cat(a, " of ", nnl, "\r")
	}
    minsABs
    }
    
sumsAB = function(CCs, verbose){
    ## calculates the length-minimum of all complexes vs each other - needed by joinCC
    sumsABs = matrix(nrow=length(CCs), ncol=length(CCs))
    nnl = length(CCs)
    lengths = NA
	for (a in 1:length(CCs))
	    lengths[a] = length(CCs[[a]])
    if (verbose) cat("\nCCs-minimums: \n")
    for (a in 1:(nnl-1)) {
	for (b in (a+1):nnl)
	    sumsABs[a,b] = sum(lengths[a],lengths[b])
	if (verbose) cat(a, " of ", nnl, "\r")
	}
    sumsABs
    }    

removeNAs = function(CCs){ ## used by joinCC to remove NAs from core complex sets
    nCCs = list()
    for (a in 1:length(CCs))
	if (!is.na(CCs[[a]][1]))
	    nCCs[[length(nCCs)+1]] = CCs[[a]]
    nCCs
    }
    
joinCC = function(coreComplexes, ovT, overlaps=NA, minsABs = NA, fromFile=T, minLength=2, jaccard=FALSE, onlyJoinable=F, verbose=F){ 
    ## joines complexes with a overlap bigger than ovT
    ## definition overlap: percentage of the smaller set must be in the bigger set or jaccard index
    ## first looks for every pair of clusters if they would join or not. in a second step, join everything
    ## fromFile: when T and overlaps and minsABs are not NA, they must contain filepaths to the matrices. when F, the contain the matrices itself
    library(RBGL)
    if (is.na(overlaps[1])){
	nCC = removeNAs(coreComplexes)
	}else{
	nCC = coreComplexes
	}
    nnl = length(nCC)
    if (nnl == 0) {
        cat("no CCs .. nothing to do.")
        return(list(unJoined = NA, joined = NA, hasJoined=FALSE))
        }
	if (nnl == 1){
		cat("only one CC .. nothing to do.")
		return(list(unJoined = nCC, joined = nCC, hasJoined=TRUE))
		}
	if (is.na(overlaps[1]))
	    overlaps = overlap(nCC, verbose=verbose)
	    else
	    if (fromFile) 
		load(file = overlaps)
	    
	if (!class(minsABs) %in% c("matrix", "character")) {
	    if (!jaccard) minsABs = minsAB(nCC, verbose=verbose) else minsABs = sumsAB(nCC, verbose=verbose) - overlaps
	    }else{
	    if (fromFile) {
		load(file = minsABs)
		minsABs =  minsABs - overlaps
		}
	    }    
	if (verbose) cat("\n")
	gc()
	be1 = which((overlaps / minsABs) > ovT)
	remove(overlaps)
	remove(minsABs)
	gc()
	adjMat = matrix(nrow=length(nCC), ncol=length(nCC), data=0)
	adjMat[be1] = 1
        adjMat = adjMat + t(adjMat)
        gc()
        rownames(adjMat) = 1:nrow(adjMat)
        colnames(adjMat) = 1:nrow(adjMat)
        gc()
   if (sum(adjMat)==0) {
        if (verbose) cat("no sufficient Overlap anywhere .. nothing to join.\n")
        return(list(unJoined = nCC, joined = nCC, hasJoined=FALSE))
        }
    aloneCCs = unique(which(rowSums(adjMat)==0)) ## not all CCs appear in the edgeslist. they get lost when i am not aware.
    CCnew = list()
    if (verbose) 
		cat("\nusing ", length(which(rowSums(adjMat)>0)), " nodes and ", sum(adjMat), "edges.")
    G2 = new("graphAM", adjMat = adjMat)
    if (verbose) 
		cat("\n Calc. Connected components:")
    if (length(which(adjMat==0))==nrow(adjMat)){ 
	CCnew[[length(CCnew)+1]] = c(1:nrow(adjMat))
	}else{
	components = connectedComp(as(G2, "graphNEL"))
	##components = connectedComp(G2)
	for (a in 1:length(components))
	    if (length(components[[a]])>1)
		CCnew[[length(CCnew)+1]] = as.numeric(components[[a]])
	}
	
	if (verbose) 
	    cat(length(CCnew), "components found. \n")
    cc2 = list()
    for (a in 1:length(CCnew)){
        prots=NULL
        for (b in 1:length(CCnew[[a]])) prots = unique(c(prots, nCC[[CCnew[[a]][b]]]))
        cc2[[a]] = prots
        }
    if (!onlyJoinable) {
        lcc2 = length(cc2)
        if (length(aloneCCs)>0)
            for (a in 1:length(aloneCCs))
                cc2[[lcc2+a]] = nCC[[aloneCCs[a]]]
            }
    cc3=list()
    for (a in 1:length(cc2))
        if (length(cc2[[a]])>=minLength) cc3[[(length(cc3)+1)]] = cc2[[a]]
	if (verbose) 
		cat(length(nCC)," coreComplexes were joined to ", length(cc3),".\n", sep="")
	return(list(unJoined = nCC, joined = cc3, hasJoined=TRUE))
    }

joinJC = function(JC, ovT, overlaps=NA, minsABs=NA, minLength=2, jaccard=FALSE, onlyJoinable=F, verbose=F){
    ## joines complexes which were already joined and occur as something$joined list
    ## joines complexes with a overlap bigger than ovT
    ## definition overlap: percentage of the smaller set must be in the bigger set or jaccard
    ## first looks for every pair of clusters if they would join or not. in a second step, join everything
    library(RBGL)
    joined = JC$joined
    nnl = length(joined)
    if (nnl == 0)
    if (nnl == 1){
	cat("only one complex .. nothing to do.")
	return(list(unJoined = joined, joined = joined, hasJoined=FALSE))
	}
	if (is.na(overlaps[1][1])) 
	    overlaps = overlap(joined, verbose=verbose)
	    else
	    load(file = overlaps)
	    
	if (is.na(minsABs[1][1])){
	    if (!jaccard) minsABs = minsAB(joined, verbose=verbose) else minsABs = sumsAB(joined, verbose=verbose) - overlaps
	    }else{
	    load(file = minsABs)
	    minsABs =  minsABs - overlaps
	    }
	if (verbose) cat("\n")
	gc()
	be1 = which((overlaps / minsABs) > ovT)
	remove(overlaps)
	remove(minsABs)
	gc()
	adjMat = matrix(nrow=length(joined), ncol=length(joined), data=0)
	adjMat[be1] = 1
        adjMat = adjMat + t(adjMat)
        gc()
        rownames(adjMat) = 1:nrow(adjMat)
        colnames(adjMat) = 1:nrow(adjMat)
        gc()
   if (sum(adjMat)==0) {
        if (verbose) cat("no sufficient Overlap anywhere .. nothing to join.\n")
        return(list(unJoined = joined, joined = joined, hasJoined=F))
        }
    
    aloneCCs = unique(which(rowSums(adjMat)==0)) ## not all CCs appear in the edgeslist. they get lost when i am not aware.
    CCnew = list()
    if (verbose) 
		cat("\nusing ", length(which(rowSums(adjMat)>0)), " nodes and ", sum(adjMat), "edges.")
    gc()
    G2 = new("graphAM", adjMat = adjMat)
    if (verbose) 
		cat("\n Calc. Connected components:")
    if (length(which(adjMat==0))==nrow(adjMat)){ 
	CCnew[[length(CCnew)+1]] = c(1:nrow(adjMat))
	}else{
	remove(adjMat)
	gc()
	components = connectedComp(as(G2, "graphNEL"))
	##components = connectedComp(G2)
	for (a in 1:length(components))
	    if (length(components[[a]])>1)
		CCnew[[length(CCnew)+1]] = as.numeric(components[[a]])
	}
	if (verbose) 
	    cat(length(CCnew), "components found. \n")
    cc2 = list()
    for (a in 1:length(CCnew)){
        prots=NULL
        for (b in 1:length(CCnew[[a]])) prots = unique(c(prots, joined[[CCnew[[a]][b]]]))
        cc2[[a]] = prots
        }
    if (!onlyJoinable) {
        lcc2 = length(cc2)
        if (length(aloneCCs)>0)
            for (a in 1:length(aloneCCs))
                cc2[[lcc2+a]] = joined[[aloneCCs[a]]]
            }
    cc3=list()
    for (a in 1:length(cc2))
        if (length(cc2[[a]])>=minLength) cc3[[(length(cc3)+1)]] = cc2[[a]]
	if (verbose) 
		cat(length(joined)," coreComplexes were joined to ", length(cc3),".\n", sep="")
	return(list(unJoined = joined, joined = cc3, hasJoined=TRUE))
    }

    

exportTree = function(HC4Nres, rnames, outfilename){
    ## Creates a hierarchical network for cytoscape.
    ## each node is a protein complex, attributes are their level and the U*
    whichOnes = 1:length(HC4Nres$hasU)
    JC = getJC(rnames, HC4Nres$HC4Nres)
    sink(paste(outfilename, "_nodes.txt", sep=""))
    cat("Node\tLevel\tCo_occurrence\tType\tProteins\n")
    for (a in whichOnes){
	cat(a, "\t", HC4Nres$comesFromLevel[a], "\t", HC4Nres$hasU[a],"\t", HC4Nres$leaf[a], "\t[", sep="")
	for (b in 1:length(JC$joined[[a]])){
	    cat("\"", rnames[JC$joined[[a]][b]],"\"", sep="")
	    if (b < length(JC$joined[[a]])) cat(",")
	    }
	cat("]")
	if (a != whichOnes[length(whichOnes)]) cat("\n")
	}
    sink()
    sink(paste(outfilename, "_edges.txt", sep=""))
    cat("Node1\tNode2\n")
    for (a in whichOnes){
	cat(a, "\t", getParent(HC4Nres$HC4Nres, a), sep="")
	if (a != whichOnes[length(whichOnes)]) cat("\n")
	}
    sink()
    }

## based on "heatmap" by Andy Liaw, original: R. Gentleman, M. Maechler, W. Huber
## performs all calculations without plotting the result.
heatmap_noplot = function (x, Rowv = NULL, Colv = if (symm) "Rowv" else NULL, 
    distfun = dist, hclustfun = hclust, reorderfun = function(d, 
        w) reorder(d, w), add.expr, symm = FALSE, revC = identical(Colv, 
        "Rowv"), scale = c("row", "column", "none"), na.rm = TRUE, hminPerc = 0) {
    ## hminPerc is added: sets hmin for the dendrogram to a min percentage of the distmatrix to avoid too deep recursion in the dendrogram
    scale <- if (symm && missing(scale)) 
        "none"
    else match.arg(scale)
    if (length(di <- dim(x)) != 2 || !is.numeric(x)) 
        stop("'x' must be a numeric matrix")
    nr <- di[1L]
    nc <- di[2L]
    if (nr <= 1 || nc <= 1) 
        stop("'x' must have at least 2 rows and 2 columns")
    doRdend <- !identical(Rowv, NA)
    doCdend <- !identical(Colv, NA)
    if (!doRdend && identical(Colv, "Rowv")) 
        doCdend <- FALSE
    if (is.null(Rowv)) 
        Rowv <- rowMeans(x, na.rm = na.rm)
    if (is.null(Colv)) 
        Colv <- colMeans(x, na.rm = na.rm)
    if (doRdend) {
        if (inherits(Rowv, "dendrogram")) 
            ddr <- Rowv
        else {
            dists = distfun(x)
            hcr <- hclustfun(dists)
            ddr <- as.dendrogram(hcr)
            if (!is.logical(Rowv) || Rowv) 
                ddr <- reorderfun(ddr, Rowv)
        }
        if (nr != length(rowInd <- order.dendrogram(ddr))) 
            stop("row dendrogram ordering gave index of wrong length")
    }
    else rowInd <- 1L:nr
    if (doCdend) {
        if (inherits(Colv, "dendrogram")) 
            ddc <- Colv
        else if (identical(Colv, "Rowv")) {
            if (nr != nc) 
                stop("Colv = \"Rowv\" but nrow(x) != ncol(x)")
            ddc <- ddr
        }
        else {
        dists = distfun(if (symm) 
            x
            else t(x))
            hcc <- hclustfun(dists)
            ddc <- as.dendrogram(hcc)
            if (!is.logical(Colv) || Colv) 
                ddc <- reorderfun(ddc, Colv)
        }
        if (nc != length(colInd <- order.dendrogram(ddc))) 
            stop("column dendrogram ordering gave index of wrong length")
    }
    else colInd <- 1L:nc
    x <- x[rowInd, colInd]
    labRow=NULL
    labCol=NULL
    labRow <- if (is.null(labRow)) 
        if (is.null(rownames(x))) 
            (1L:nr)[rowInd]
        else rownames(x)
    else labRow[rowInd]
    labCol <- if (is.null(labCol)) 
        if (is.null(colnames(x))) 
            (1L:nc)[colInd]
        else colnames(x)
    else labCol[colInd]
    invisible(list(rowInd = rowInd, colInd = colInd))
}
